import cv2
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import axes3d
from matplotlib import cm
from numba import jit

@jit
def visualizeHueSatHistogram(hist, disp_size, is_legend = False):
    max_hist = np.max(hist)
    width = disp_size[0]
    height = disp_size[1]
    hbin = hist.shape[0]
    sbin = hist.shape[1]

    disp = np.zeros((height, width, 3), np.uint8)

    for j in range(height):
        for i in range(width):
            hue = (i * 180.0) / width
            sat = (j * 256.0) / height
            if is_legend:
                value = 255
            else:
                value = int(hist[int(hue * hbin / 180.0), int(sat * sbin / 256.0)] / max_hist * 255.0)
            disp[j, i] = (int(hue), int(sat), value)

    return disp

@jit
def createHueSatHistogram(image, hist_size, is_weighted = True):
    width = image.shape[1]
    height = image.shape[0]
    w_normalizer = 2.0 / width
    h_normalizer = 2.0 / height
    cx = int(width / 2)
    cy = int(height / 2)

    hbin = hist_size[0]
    sbin = hist_size[1]
    hist = np.zeros((hbin, sbin))

    for j in range(height):
        for i in range(width):
            if is_weighted:
                x_norm = (i - cx) * w_normalizer
                y_norm = (j - cy) * h_normalizer
                rr = x_norm * x_norm + y_norm * y_norm
                weight = 1 - rr if (rr <= 1) else 0.0
            else:
                weight = 1.0
            hue = image[j, i, 0]
            sat = image[j, i, 1]
            hist[int(hue * hbin / 180.0), int(sat * sbin / 256.0)] += weight

    hist_sum = np.sum(hist)
    hist /= hist_sum
    return hist


if __name__ == '__main__':
    cv2.namedWindow('input')
    cv2.namedWindow('histogram')

    #cap = cv2.VideoCapture('books.mp4')
    #cap = cv2.VideoCapture('vtest.avi')
    cap = cv2.VideoCapture(1)

    try:
        hbin = 64
        sbin = 32
        hist_size = (hbin, sbin)
        hist_weighted = False
        hist_disp_size = (256, 256)

        fig = plt.figure()
        ax = fig.gca(projection='3d')

        while True:
            retval, input_bgr = cap.read()
            #input_bgr = cv2.imread('baboon.jpg')

            input_hsv = cv2.cvtColor(input_bgr, cv2.COLOR_BGR2HSV)
            hist = createHueSatHistogram(input_hsv, hist_size, hist_weighted)

            hist_vis = visualizeHueSatHistogram(hist, hist_disp_size, False)
            hist_legend = visualizeHueSatHistogram(hist, hist_disp_size, True)
            hist_vis = cv2.hconcat([hist_legend, hist_vis])

            cols = np.arange(hbin)
            rows = np.arange(sbin)
            mesh_rows, mesh_cols = np.meshgrid(rows, cols)
            ax.plot_surface(mesh_rows, mesh_cols, hist, cmap=cm.coolwarm)
            ax.set_xlabel('sat')
            ax.set_ylabel('hue')
            plt.pause(0.001)
            ax.clear()

            cv2.imshow("input", input_bgr)
            cv2.imshow("histogram", cv2.cvtColor(hist_vis, cv2.COLOR_HSV2BGR))
            key = cv2.waitKey(30)
            if key == ord('q'):
                break

    finally:
        cap.release()
        cv2.destroyAllWindows()
