import math
import cv2
import numpy as np
from numpy.linalg import inv
from numba import jit

import color_histogram as ch

@jit
def meanShiftVector(image_area, hist_model, hist_cand):
    width =  image_area.shape[1]
    height =  image_area.shape[0]
    cx = (width - 1) / 2.0
    cy = (height - 1) / 2.0
    hbin = hist_model.shape[0]
    sbin = hist_model.shape[1]
    w_normalizer = 2.0 / width
    h_normalizer = 2.0 / height

    m0 = mx = my = 0.0
    for j in range(height):
        for i in range(width):
            hue = image_area[j, i, 0]
            sat = image_area[j, i, 1]
            q = hist_model[int(hue * hbin / 180.0), int(sat * sbin / 256.0)]
            p = hist_cand[int(hue * hbin / 180.0), int(sat * sbin / 256.0)]

            x_norm = (i - cx) * w_normalizer
            y_norm = (j - cy) * h_normalizer
            rr = x_norm * x_norm + y_norm * y_norm

            if rr < 1 and p != 0.0:
                w = math.sqrt(q / p)
                m0 += w
                mx += w * i
                my += w * j
    if m0 == 0:
        return 0.0, 0.0
    else:
        return mx/m0 - cx, my/m0 - cy
    

def trackByMeanShift(image, tcenter, tsize, hist_model, hist_size):
    iwidth = image.shape[1]
    iheight = image.shape[0]
    twidth = tsize[0]
    theight = tsize[1]
    w_normalizer = 2.0 / twidth
    h_normalizer = 2.0 / theight
    half_w = int(twidth / 2)
    half_h = int(theight / 2)
    hbin = hist_size[0]
    sbin = hist_size[1]
    max_iter = 20

    for iter in range(max_iter):
        candidate = cv2.getRectSubPix(image, (twidth, theight), tcenter)
        hist_cand = ch.createHueSatHistogram(candidate, hist_size, True)
        mv_x, mv_y = meanShiftVector(candidate, hist_model, hist_cand)
        tcenter = (tcenter[0] + mv_x, tcenter[1] + mv_y)
        dd = mv_x * mv_x + mv_y * mv_y 
        if dd < 1.0:
            break

    bhattacharyya_coeff = np.sum(np.sqrt(hist_model * hist_cand))
        
    cv2.imshow("hist_cand", cv2.cvtColor(
        ch.visualizeHueSatHistogram(hist_cand, (256, 256), False), 
        cv2.COLOR_HSV2BGR))

    return tcenter, bhattacharyya_coeff


def onMouse(event, x, y, flag, param):
    if track_state == "selecting":
        param[2] = (x, y)

    if event == cv2.EVENT_LBUTTONDOWN:
        param[0] = "selecting"
        param[1] = (x, y)
    elif event == cv2.EVENT_LBUTTONUP:
        if mouse_pos2[0] > mouse_pos1[0] and mouse_pos2[1] > mouse_pos1[1]:
            param[0] = "to_be_initialized"
        else:
            param[0] = "off"


if __name__ == '__main__':
    cv2.namedWindow('track')
    #cap = cv2.VideoCapture('books.mp4')
    cap = cv2.VideoCapture('vtest.avi')
    #cap = cv2.VideoCapture(1)

    try: 
        hbin = 64
        sbin = 64
        hist_size = (hbin, sbin)

        print("hit p key to toggle pausing")
        print("drag to set the model region")

        # track_state, mouse_pos1 (top-left), mouse_pos2 (bottom-right)
        mouse_param = ["off", (-1, -1), (-1, -1)]
        cv2.setMouseCallback('track', onMouse, mouse_param)

        is_paused = False
    
        while True:
            if not is_paused:
                retval, input_bgr = cap.read()
            input_disp = input_bgr.copy()
            input_hsv = cv2.cvtColor(input_bgr, cv2.COLOR_BGR2HSV)
    
            track_state = mouse_param[0]
            mouse_pos1 = mouse_param[1]
            mouse_pos2 = mouse_param[2]

            if track_state == "to_be_initialized":
                tcenter = (int((mouse_pos2[0] + mouse_pos1[0]) / 2), 
                           int((mouse_pos2[1] + mouse_pos1[1]) / 2))
                twidth = int((mouse_pos2[0] - mouse_pos1[0]))
                theight = int((mouse_pos2[1] - mouse_pos1[1]))
                tsize = (twidth, theight)
                
                model = cv2.getRectSubPix(input_hsv, tsize, tcenter)
                hist_model = ch.createHueSatHistogram(model, hist_size, True)
                mouse_param[0] = "tracking"
                cv2.imshow("model", cv2.cvtColor(model, cv2.COLOR_HSV2BGR))
                cv2.imshow("hist_model", cv2.cvtColor(
                    ch.visualizeHueSatHistogram(hist_model, (256, 256), False),
                    cv2.COLOR_HSV2BGR))
    
            elif track_state == "tracking":
                tcenter, bc = trackByMeanShift(input_hsv, tcenter, tsize, 
                                               hist_model, hist_size)
                cv2.ellipse(input_disp, (int(tcenter[0]), int(tcenter[1])),
                            (int(twidth / 2), int(theight / 2)), 
                            0, 0, 360, (0, 0, 255), 3)

            elif (track_state == "selecting" 
                  and mouse_pos2[0] > mouse_pos1[0] 
                  and mouse_pos2[1] > mouse_pos1[1]):

                cv2.ellipse(input_disp, 
                            (int((mouse_pos1[0] + mouse_pos2[0]) / 2), 
                             int((mouse_pos1[1] + mouse_pos2[1]) / 2)), 
                            (int((mouse_pos2[0] - mouse_pos1[0] + 1) / 2), 
                             int((mouse_pos2[1] - mouse_pos1[1] + 1) / 2)), 
                            0, 0, 360, (0, 255, 0), 1)

            cv2.imshow("track", input_disp)
            key = cv2.waitKey(30)
            if key == ord('q'):
                break
            elif key == ord('p'):
                is_paused = not is_paused

    finally:
        cap.release()
        cv2.destroyAllWindows()
